
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
 
entity testbench is
	generic
	(
		address_width_testbench : integer := 3;
		data_width_testbench    : integer := 4
	);
end testbench;
 
architecture testbench_rtl of testbench is 
 
    component cam
	 generic (
					address_width_cam : integer := 3;
					data_width_cam    : integer := 4
				);
    port(
			------------------ Inputs
         clock_cam 			: in  	std_logic;
         data_in_cam 		: in  	std_logic_vector(data_width_cam-1 downto 0);
         write_enable_cam 	: in  	std_logic;
         write_address_cam : in  	std_logic_vector(address_width_cam-1 downto 0);
         read_enable_cam 	: in  	std_logic;
         read_address_cam 	: in  	std_logic_vector(address_width_cam-1 downto 0);
         lookup_cam 			: in  	std_logic;
			------------------ Outputs
         data_out_cam 		: out  	std_logic_vector(data_width_cam-1 downto 0);
         address_out_cam 	: out  	std_logic_vector(address_width_cam-1 downto 0);
         error_cam 			: out  	std_logic
        );
    end component;
    

   --Inputs
   signal clock_cam 				: std_logic := '0';
	signal data_in_cam 			: std_logic_vector(data_width_testbench-1 downto 0) := (others => '0');
   signal write_enable_cam 	: std_logic := '0';
   signal write_address_cam 	: std_logic_vector(address_width_testbench-1 downto 0) := (others => '0');
   signal read_enable_cam 		: std_logic := '0';
   signal read_address_cam 	: std_logic_vector(address_width_testbench-1 downto 0) := (others => '0');
   signal lookup_cam 			: std_logic := '0';

 	--Outputs
   signal data_out_cam 			: std_logic_vector(data_width_testbench-1 downto 0);
   signal address_out_cam 		: std_logic_vector(address_width_testbench-1 downto 0);
   signal error_cam 				: std_logic;

   -- Clock period definitions
   constant clock_cam_period : time := 20ns;
	-- Address and data definitions
	signal write_address, read_address	: natural range 2 ** address_width_testbench-1 downto 0 := 0;
	signal data_in	: natural range data_width_testbench-1 downto 0 := 0;
	-- Type string has a positive range
	signal test_string : string (1 to 43) := "The quick brown fox jumps over the lazy dog";
	signal char : character;
	signal start_writing, start_reading, start_lookup : std_logic := '0';
	-- The different states for the processes
	type write_states is (idle_state, init_write_state, write_state, stop);
	type read_states is (idle_state, init_read_state, read_state, stop);
	type lookup_states is (idle_state, init_lookup_state, lookup_state, stop);
	type testbench_states is (writing, reading, lookup, stop);
	-- The state definitions
	signal state_read : read_states := idle_state;
	signal state_write : write_states := idle_state;
	signal state_lookup : lookup_states := idle_state;
	signal state_testbench : testbench_states := writing;
begin
   uut: cam port map 
	(
			--------------- Inputs
          clock_cam 				=> clock_cam,
          data_in_cam 			=> data_in_cam,
          write_enable_cam 	=> write_enable_cam,
          write_address_cam 	=> write_address_cam,
          read_enable_cam 		=> read_enable_cam,
          read_address_cam 	=> read_address_cam,
          lookup_cam 			=> lookup_cam,
			 -------------- Outputs
          data_out_cam 			=> data_out_cam,
          address_out_cam 		=> address_out_cam,
          error_cam 				=> error_cam
    );

   -- Clock process definitions
   clock_cam_process :process
   begin
		clock_cam <= '0';
		wait for clock_cam_period/2;
		clock_cam <= '1';
		wait for clock_cam_period/2;
   end process;
 

   write_to_cam : process (clock_cam)		
   begin		
     if rising_edge(clock_cam)
	  then
				case state_write is
					when idle_state =>			if start_writing = '1'
														then
															state_write <= init_write_state;
														else
															null;
														end if;
					when init_write_state => 	write_address 					<= 0;
														data_in							<= 0;
														write_enable_cam 				<= '1';
														state_write 					<= write_state;
					when write_state =>			if write_address < 2**address_width_testbench
														then
															-- char <= test_string(write_address + 1);
															-- Convert character to std_logic_vector
															-- data_in_cam <= std_logic_vector(to_unsigned(character'pos(char),data_width_testbench));
															write_address_cam <= std_logic_vector(to_unsigned(write_address,address_width_testbench));														
															data_in_cam <= std_logic_vector(to_unsigned(data_in,data_width_testbench));
															-- Increment address
															write_address <= write_address+1;
															-- Increment data
															data_in <= data_in + 1;
														else
															state_write <= stop;
														end if;
					when stop =>	 				write_enable_cam 	<= '0';
														-- start_writing 		<= '0';
				end case;
     end if; 
   end process;
	
	read_from_cam : process (clock_cam)
		variable char : character;
	   variable address : natural;
	begin
		if rising_edge(clock_cam)
		then
				case state_read is
					when idle_state =>		if start_reading = '1'
													then
														state_read <= init_read_state;
													else
														null;
													end if;
					when init_read_state => read_address 		<= 0;
													read_enable_cam 	<= '1';
													state_read 			<= read_state;
					when read_state =>		--if address < test_string'length
													if read_address < 2 ** address_width_testbench
													then
														-- Convert integer to std_logic_vector
														read_address_cam <= std_logic_vector(to_unsigned(read_address,address_width_testbench));
														-- Convert std_logic_vector to character
													   -- char := character'val(to_integer(unsigned(data_out_cam)));
														-- Compare the char to the char in the test_string
														-- assert char = test_string(address+1);
														-- Increment address
														read_address <= read_address + 1;
													else
														state_read <= stop;													
													end if;
					when stop =>				read_enable_cam 	<= '0';
													-- start_reading 		<= '0';
				end case;
			
		end if;
   end process;
	
	lookup_at_cam : process (clock_cam)
		variable char : character;
	begin
		if rising_edge(clock_cam)
		then		
				case state_lookup is
					when idle_state =>			if start_lookup = '1'
														then
															state_lookup <= init_lookup_state;
														else
															null;
														end if;
					when init_lookup_state => 	lookup_cam 			<= '1';
														-- Lookup for known char
														-- char := 'T';
														-- Convert character to std_logic_vector
														-- data_in_cam <= std_logic_vector(to_unsigned(character'pos(char),data_width_testbench));
														data_in_cam <= "0011";
														state_lookup <= lookup_state;
					when lookup_state =>			assert error_cam = '0';
														-- Lookup for unknown char
														-- char := '6';
														-- Convert character to std_logic_vector
														-- data_in_cam <= std_logic_vector(to_unsigned(character'pos(char),data_width_testbench));
														assert error_cam = '1';
														state_lookup <= stop;
					when stop =>					lookup_cam 		<= '0';
														-- start_lookup 	<= '0';
				end case;
		end if;
	end process;
	
	control : process (clock_cam)		
	begin	
		if rising_edge (clock_cam)
		then
		   case state_testbench is
				when writing => 	start_writing <= '1';
										if state_write = stop
										then
											state_testbench <= reading;
										else
											null;
										end if;
				when reading => 	start_writing 		<= '0';
										start_reading <= '1';
										if state_read = stop
										then
										  state_testbench <= lookup;
										else
											null;
										end if;
				when lookup => 	start_reading 		<= '0';
										start_lookup 		<= '1';
										if state_lookup = stop
										then
											state_testbench <= stop;
										else
											null;
										end if;
				when stop =>		-- start_lookup <= '0';
			end case;
		end if;
	end process;
	
end architecture testbench_rtl;
